/* -*-Asm-*- */
/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 1996  Erich Boleyn  <erich@uruk.org>
 *  Copyright (C) 1999, 2000, 2001  Free Software Foundation, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stage1.h>
	
/*
 *  defines for the code go here
 */

	/* Absolute addresses
	   This makes the assembler generate the address without support
	   from the linker. (ELF can't relocate 16-bit addresses!) */
#define ABS(x) (x-_start+0x7c00)

	/* Print message string */
#define MSG(x)	movw $ABS(x), %si; call message

	.file	"stage1.S"

	.text

	/* Tell GAS to generate 16-bit instructions so that this code works
	   in real mode. */
	.code16

.globl _start; _start:
	/*
	 * _start is loaded at 0x7c00 and is jumped to with CS:IP 0:0x7c00
	 */

	/*
	 * Beginning of the sector is compatible with the FAT/HPFS BIOS
	 * parameter block.
	 */

	jmp	after_BPB
	nop	/* do I care about this ??? */

	/*
	 * This space is for the BIOS parameter block!!!!  Don't change
	 * the first jump, nor start the code anywhere but right after
	 * this area.
	 */

	. = _start + 3

	/* OEM identifier */
	.string	"GRUB 0.5"


	. = _start + STAGE1_BPBEND

	/*
	 * End of BIOS parameter block.
	 */

stage1_version:	
	.byte	COMPAT_VERSION_MAJOR, COMPAT_VERSION_MINOR
boot_drive:	
	.byte 0xff	/* the disk to load stage2 from */
			/* 0xff means use the boot drive */
flags:
	.byte	0
stage2_address:
	.word	0x8000
stage2_sector:
	.long	1

after_BPB:

/* general setup */
	cli		/* we're not safe here! */

	/*
	 * ljmp to the next instruction because some bogus BIOSes
	 * jump to 07C0:0000 instead of 0000:7C00.
	 */
	ljmp	$0, $ABS(real_start)

real_start:	

	/* set up %ds, %es and %ss as offset from 0 */
	xorw	%ax, %ax
	movw	%ax, %ds
	movw	%ax, %es
	movw	%ax, %ss

	/* set up the REAL stack */
	movw	$STAGE1_STACKADDR, %sp

	sti		/* we're safe again */

	/* set %di to the stage1 address */
	movw	$0x7c00, %di

	/*
	 *  Check if we have a forced disk reference here
	 */
	movb	STAGE1_BOOT_DRIVE(%di), %al
	cmpb	$0xff, %al
	je	1f
	movb	%al, %dl
1:
	/* save drive reference! */
	movb	%dl, STAGE1_BOOT_DRIVE(%di)

	/* print a notification message on the screen */
	MSG(notification_string)

	/* do not probe LBA if the drive is a floppy */
	orb	%dl, %dl
	jns	chs_mode
			
	/* check if LBA is supported */
	movb	$0x41, %ah
	movw	$0x55aa, %bx
	int	$0x13

	/* use CHS if fails */
	jc	chs_mode
	cmpw	$0xaa55, %bx
	jne	chs_mode

	/* check if AH=0x42 is supported if FORCE_LBA is zero */
	testb	$STAGE1_FLAG_FORCE_LBA, STAGE1_FLAGS(%di)
	jnz	skip_lba_bitmap_check
	andw	$1, %cx
	jz	chs_mode
	
skip_lba_bitmap_check:
	
	/* get the geometry (limited to 2TB!) */
	movb	$0x48, %ah
	movw	$STAGE1_DRP_ADDR, %si
	movw	$STAGE1_DRP_SIZE, (%si)
	int	$0x13

	/* use CHS if fails */
	jc	chs_mode

lba_mode:
	/* save the total number of sectors */
	movl	0x10(%si), %ecx

	/* set %si to the disk address packet */
	movw	%di, %si

	/* set the mode to non-zero */
	movb	$1, (%si)
	incw	%si
	
	/* the size and the reserved byte */
	movw	$0x0010, (%si)

	/* the blocks */
	movw	$1, 2(%si)
	
	/* the segment of buffer address */
	movw	STAGE1_STAGE2_ADDRESS(%di), %bx
	movw	%bx, 4(%si)

	xorl	%eax, %eax
	movw	%ax, 6(%si)
	movl	%eax, 12(%si)

	movl	STAGE1_STAGE2_SECTOR(%di), %ebx
	/* check for the geometry */
#ifdef NO_BUGGY_BIOS_IN_THE_WORLD
	cmpl	%ecx, %ebx
	jge	geometry_error
#endif /* NO_BUGGY_BIOS_IN_THE_WORLD */

	/* the absolute address (low 32 bits) */	
	movl	%ebx, 8(%si)
	
/*
 * BIOS call "INT 0x13 Function 0x42" to read sectors from disk into memory
 *	Call with	%ah = 0x42
 *			%dl = drive number
 *			%ds:%si = segment:offset of disk address packet
 *	Return:
 *			%al = 0x0 on success; err code on failure
 */

	movb	$0x42, %ah
	int	$0x13

	jnc	boot_stage2
	/* LBA read is not supported, so fallback to CHS.  */

chs_mode:	
	/*
	 *  Check if we should use the information in the bpb
	 *  We use it if force_geometry is set and bpb_drive equals
	 *  the stage2 drive.
	 */
	testb	$STAGE1_FLAG_FORCE_BPB, STAGE1_FLAGS(%di)
	jz	probe_geometry
	cmpb	STAGE1_BPB_DRIVE(%di), %dl
	jnz	probe_geometry

use_bpb_geometry:

	xorl	%edx, %edx
	orw	STAGE1_BPB_TOTAL16(%di), %dx
	jnz	1f
	movl	STAGE1_BPB_TOTAL32(%di), %edx
1:	
	movw	STAGE1_BPB_SECTORS(%di), %cx
	movzwl	STAGE1_BPB_HEADS(%di), %eax
	mulb	%cl
	movw	%ax, %bx	/* %bx = sectors per cylinder */
	shll	$10, %eax	/* %eax = total sectors to 1024 cyl boundary */

	cmpl	%edx, %eax
	jae	1f
	movl	%eax, %edx
1:	movl	%edx, STAGE1_CHS_SEC_DISK(%di)

store_geometry:	
	/* set the mode to zero */
	movb    $0, (%di)
	movw	%cx, STAGE1_CHS_SEC_TRACK(%di)
	movw	%bx, STAGE1_CHS_SEC_CYL(%di)
/*
 *  Take care of BIOS geometry translation (ugh!)
 */

	/* load logical sector start (bottom half) */
	movl	STAGE1_STAGE2_SECTOR(%di), %eax
	movw	STAGE1_STAGE2_SECTOR+2(%di), %dx
	cmpl	STAGE1_CHS_SEC_DISK(%di), %eax
	jae	geometry_error

	/* divide by number of sectors per cylinder */
	divw	%bx

	/* cylinders to %dx, sector and head to %ax */
	xchgw	%dx, %ax

	/* divide by number of sectors per track */
	divb	%cl
	movb	%ah, %cl
	incb	%cl		/* normalize sector (sectors go
					from 1-N, not 0-(N-1) ) */
	shlb	$6, %dh
	movb	%dl, %ch
	orb	%dh, %cl	/* composite together */

	movb	%al, %dh	/* head number */
	movb	STAGE1_BOOT_DRIVE(%di), %dl

/*
 * BIOS call "INT 0x13 Function 0x2" to read sectors from disk into memory
 *	Call with	%ah = 0x2
 *			%al = number of sectors
 *			%ch = cylinder
 *			%cl = sector (bits 6-7 are high bits of "cylinder")
 *			%dh = head
 *			%dl = drive (0x80 for hard disk, 0x0 for floppy disk)
 *			%es:%bx = segment:offset of buffer
 *	Return:
 *			%al = 0x0 on success; err code on failure
 */

	movw	STAGE1_STAGE2_ADDRESS(%di), %bx
	movw	$0x0201, %ax	/* function 2, 1 sector */
	int	$0x13

	movw	$ABS(read_error_string), %si
	jc	general_error
boot_stage2:
	/* boot stage2 */
	jmp	*(stage2_address)

/* Error section -- This section is in the middle, so that we can use
 * short conditional jumps.
 */

/*
 * BIOS Geometry translation error (past the end of the disk geometry!).
 */
geometry_error:
	movw	$ABS(geometry_error_string), %si

general_error:
	call	message
	movw	$ABS(general_error_string), %si
	call	message

/* go here when you need to stop the machine hard after an error condition */
stop:	jmp	stop

/* End of Error section.
 */

probe_geometry:	
	/*
	 *  Determine the hard disk geometry from the BIOS!
	 *  We do this first, so that LS-120 IDE floppies work correctly.
	 */
	pushw	%di
	movb	$8, %ah
	int	$0x13
	popw	%di
	jc	probe_floppy_or_die

final_init:
	/* calculate number of heads */
	xorw	%ax, %ax
	movb	%dh, %al
	incw	%ax
	
	/* calculate number of cylinders */
	movb	%cl, %bh
	shrb	$6, %bh
	movb	%ch, %bl

	/* calculate number of sectors */
	andw	$0x3f, %cx
	mulw	%cx
	xchgw	%ax, %bx
	mulw	%bx
	movw	%ax, STAGE1_CHS_SEC_DISK(%di)
	movw	%dx, STAGE1_CHS_SEC_DISK+2(%di)
	jmp	store_geometry	

probe_floppy_or_die:
	/*
	 *  The probe call failed, so maybe use the floppy probe instead.
	 */
	orb	%dl, %dl
	jns	floppy_probe

	/* But if it is a hard disk we're screwed. */
	movw	$ABS(hd_probe_error_string), %si
	jmp	general_error

/*
 * message: write the string pointed to by %si
 *
 *   WARNING: trashes %si, %ax, and %bx
 */

	/*
	 * Use BIOS "int 10H Function 0Eh" to write character in teletype mode
	 *	%ah = 0xe	%al = character
	 *	%bh = page	%bl = foreground color (graphics modes)
	 */
1:
	movw	$0x0001, %bx
	movb	$0xe, %ah
	int	$0x10		/* display a byte */
message:
	lodsb
	orb	%al, %al
	jne	1b	/* if not end of string, jmp to display */
	ret


notification_string:	.string "GRUB "
geometry_error_string:	.string "Geom"
hd_probe_error_string:	.string "Hard Disk"
read_error_string:	.string "Read"
general_error_string:	.string " Error"

	/*
	 *  Windows NT breaks compatibility by embedding a magic
	 *  number here.
	 */

	. = _start + STAGE1_WINDOWS_NT_MAGIC
nt_magic:	
	.long 0
	.word 0

	/*
	 *  This is where an MBR would go if on a hard disk.  The code
	 *  here isn't even referenced unless we're on a floppy.  Kinda
	 *  sneaky, huh?
	 */

part_start:	
	. = _start + STAGE1_PARTSTART

probe_values:
	.byte	36, 18, 15, 9, 0

floppy_probe:
/*
 *  Perform floppy probe.
 */

	movw	$ABS(probe_values), %si

probe_loop:
	/* reset floppy controller INT 13h AH=0 */
	xorw	%ax, %ax
	int	$0x13

	lodsb
	movb	%al, %cl

	/* if number of sectors is 0, display error and die */
	orb	%cl, %cl
	je	fd_probe_error

	/* perform read */
	movw	$STAGE1_BUFFERSEG, %bx
	movw	$0x201, %ax
	movb	$0, %ch
	movb	$0, %dh
	int	$0x13

	/* if error, jump to "probe_loop" */
	jc	probe_loop

	/* %cl has already the correct value! */
	movb	$1, %dh		/* two heads */
	movb	$79, %ch	/* 80 tracks */

	jmp	final_init

/*
 * Floppy disk probe failure.
 */
fd_probe_error:	
	movw	$ABS(fd_probe_error_string), %si
	jmp	general_error

fd_probe_error_string:	.string "Floppy"

	. = _start + STAGE1_PARTEND

/* the last 2 bytes in the sector 0 contain the signature */
	.word	STAGE1_SIGNATURE
